(in-package nc)

(defun domain (vertex-type)
  (case vertex-type
    ((L) '((R L)   (L R)   (+ R)   (L +)   (- L)   (R -)))
    ((Y) '((+ + +) (- - -) (L R -) (- L R) (R - L)))
    ((T) '((R L +) (R L -) (R L L) (R L R)))
    ((W) '((L R +) (- - +) (+ + -)))))



(defun consistency-func (labelset1 labelset2 pos1 pos2)
	   (let* ((newlabelset2  
		   (loop for label in labelset2
		       when (find label labelset1 
				  :test #'(lambda (l2 l1)
					    (label-fit (nth pos2 l2)
						       (nth pos1 l1))))
		       collect label))
		  (newlabelset1  
		   (loop for label in labelset1
		       when (find label newlabelset2 
				  :test #'(lambda (l1 l2)
					    (label-fit (nth pos1 l1)
						       (nth pos2 l2))))
		       collect label)))
	     (list newlabelset1 newlabelset2)))

(defun label-fit (l1 l2)
  (case l1
    ((+) (eq l2 '+))
    ((-) (eq l2 '-))
    ((L) (eq l2 'R))
    ((R) (eq l2 'L))))

(defmacro defdiagram (name &rest vertex-descrs)
  "each vertex will be represented as a value-node, and the edges between the vertexes are 
represented as function-nodes."
  `(defnet ,name 
       ,(loop for vertex in  vertex-descrs
	    collect (list (first vertex)
			  :content (domain (second vertex))))
     ,(loop for vertexes on vertex-descrs
	  for vertex1 = (first vertexes)
	  append (loop for vertex2 in (rest vertexes) 
		     when (member (car vertex2)  (cddr vertex1))
		     collect (list (list (car vertex1) (car vertex2)) 
				   `(lambda (l1 l2) 
				      (consistency-func l1 l2  
							,(position (car vertex2) 
								   (cddr vertex1))
							,(position (car vertex1) 
								   (cddr vertex2))))
				     (car vertex1)
				     (car vertex2))))))


(defmethod value-new ((node value-node) value)
  (< (length value) (length (content node))))

(defmethod value-ok ((node value-node) value)
  value)


;;; this macro creates a m-stair drawing.
;;; a cube is a 0-stair drawing, and a 2-stair drawing looks like this
;;;               ------------
;;;              /           /|
;;;              ------------ |
;;;           2 |           | |
;;;             |           | |
;;;             ------------  |
;;;            /           /  |
;;;            ------------   |
;;;         1 |           |   |
;;;           |           |   |  
;;;           ------------   /
;;;          /           /  /  
;;;          ------------  /  
;;;       0 |           | /   
;;;         |           |/    
;;;         -------------
(defmacro stairs (m)
  (let ((n (1+ (* 2 m))))
    `(defdiagram ,(intern (format nil "~d-STAIRS" m) "NETCLOS")
	 (r1 L v0-0 r2)
       (r2 W r3 r1 v0-1)
       (r3 L r2 ,(vertex-name n 1))
       (v0-0 W r1 v1-0 v0-1)
       (v0-1 Y v0-0 v1-1 r2)
       ,@(recursive-stairs m)
       (,(vertex-name n 0) L ,(vertex-name n 1) ,(vertex-name (1- n) 0))
       (,(vertex-name n 1) W ,(vertex-name n 0) r3 ,(vertex-name (1- n) 1)))))
     

(defun recursive-stairs (m)
  (loop for n from 1 upto m
      append (list (list (vertex-name (1- (* 2 n)) 0) 'Y  
			 (vertex-name (* 2 (1- n)) 0)
			 (vertex-name (* 2 n) 0)
			 (vertex-name (1- (* 2 n)) 1))
		   (list (vertex-name (1- (* 2 n)) 1) 'W 
			 (vertex-name (* 2 n) 1)
			 (vertex-name (* 2 (1- n)) 1)  
			 (vertex-name (1- (* 2 n)) 0))
		   (list (vertex-name (* 2 n) 0) `W
			 (vertex-name (1- (* 2 n)) 0)
			 (vertex-name (1+ (* 2 n)) 0)
			 (vertex-name (* 2 n) 1))
		   (list (vertex-name (* 2 n) 1) 'Y
			 (vertex-name (* 2 n) 0)
			 (vertex-name (1+ (* 2 n)) 1)
			 (vertex-name (1- (* 2 n)) 1)))))

(defun vertex-name (n m)
  (intern (format nil "V~d-~d" n m) "NETCLOS"))
			 
			 
			 
#|          ->
       e-----------c
  _   /           /|
  /| /         + / | 
 /  /           /  | |
   /      +    /   | V    A cube and a possible labeling
  b------------a   |
  |            |   |
  |            |   |
 ^|           +|   f
 ||            |  /
  |            | /
  |            |/
  g------------d
     <-
(defdiagram cube
    (a Y b c d)
  (b W g e a)
  (c W e f a)
  (d W f g a)
  (e L c b)
  (f L d c)
  (g L b d))
|#






#| How to test:


NC(5): (stairs 100)
<Entering interface>
<Entering read_params>
Reading parameter modification file `/home/hotz/tools/Chaco-2.0/exec/User_Params'
  Parameter `OUTPUT_METRICS' reset to 2
  Parameter `ECHO' reset to 1
  Parameter `OUTPUT_ASSIGN' reset to True

<Entering submain>
<Entering check_input>
<Entering reflect_input>


           Input and Parameter Values

# vertices = 1016, # edges = 1218
Global method: Spectral
Local method: Kernighan-Lin
Partitioning target: 1-dimensional mesh of size 2
Partitioning mode: Bisection
Random seed: 7654321



Starting to partition ...

<Entering balance>
<Entering divide, nvtxs = 1016, nedges = 1218>
<Entering eigensolve, nvtxs = 1016, nedges = 1218>
<Entering lanczos_so>
<Entering assign, nvtxs = 1016, ndims = 1>
<Entering klspiff, nvtxs = 1016>


                     Partitioning Results

After full partitioning  (nsets = 2)

                            Total      Max/Set      Min/Set
                            -----      -------      -------
Set Size:                    1016          508          508
Edge Cuts:                      3            3            3
Mesh Hops:                      3            3            3
Boundary Vertices:              6            3            3
Boundary Vertex Hops:           6            3            3
Adjacent Sets:                  2            1            1
Internal Vertices:           1010          505          505


Total time: 0.47 sec.
  partitioning 0.45
  evaluation 0.01
  other 0.01

KL time: 0.02 sec.
  nway refinement 0.02

Lanczos time: 0.43 sec.
#<RELAXATION-NET @ #x97e10a>
NC(7): (setq *net* *)
#<RELAXATION-NET @ #x97e10a>
NC(8): (relax-net *net*)
10:42,56 Uhr
(NIL NIL)
NC(9): 


|#




