(in-package nc)
;;; you generate the .so-file by
;;; make in /home/hotz/tools/Chaco-2.0/code/ to compile chaco-source files
;;; make lib  in the same directory
;;; gcc -c -g -o /home/trowe/lisp/pvm/interface.o -fPIC /home/trowe/lisp/pvm/interface.c
;;; ld -G -o /home/trowe/lisp/pvm/interface.so /home/trowe/lisp/pvm/interface.o -lchaco -lm
(load "~trowe/lisp/pvm/interface.so")
(defvar *graph* ())
(defforeign 'interface
    :entry-point (convert-to-lang "interface")
    :arguments '(integer
		 (simple-array fixnum (*))
		 (simple-array fixnum (*))
		 (simple-array fixnum (*))
		 (simple-array single-float (*))
					;(simple-array single-float (*))
					;(simple-array single-float (*))
					;(simple-array single-float (*))
					;string
					;string
		 (simple-array fixnum (*))
		 integer
		 integer
		 (simple-array fixnum (3))
					;(simple-array double-float (*))
		 integer
		 integer
		 integer
		 integer
		 integer
		 double-float
		 integer))
		 
		 
		 
(defclass tig (active-object) 
  ((tasks :accessor tasks)))
	

(defpargeneric (setf tasks) :now (tasks tig))

(defpargeneric tasks :now (tig))

(defgeneric add-task (tig task))

(defclass task (mobile-object)
  ((size :accessor size :initarg :size)
   (connected-tasks :accessor connected-tasks :initform ())))

(defclass toy-strategy () ())

(defclass chaco-strategy () ())

(defclass mean-field-strategy () 
  ((initial-temperature :initform 5.0 
			:accessor initial-temperature)
   (current-temperature :initform 5.0
			:accessor current-temperature)
   (cooling-length :accessor cooling-length
		:initarg :cooling-length
		:documentation 
		"After cooling-length effectless iterations a new temperature will be chosen")
   (epsilon :accessor epsilon 
	    :initform 0.5
	    :documentation 
	    "An iteration is effectless, if the change in the energy-function is less than epsilon")
   (effectless-iterations :accessor effectless-iterations 
			  :initform 0
			  :documentation "The number of effectless iterations in a row")))


(defmacro compile-graph (type tasks connectivity machine &rest initargs)
  (let ((strategy (apply #'make-instance type initargs)))
    `(let ((graph (make-instance ',type ,@initargs)))
       (set-tasks graph 
		  (list ,@(loop for task in tasks 
			      for location in (compute-location strategy
								tasks 
								connectivity 
								machine)
			      collect `(make-instance ,@(first task) 
					 :location (find ,location (spaces *manager*) 
							 :key #'host)))))
       (connect graph ,connectivity)
       graph)))
				       
(defun make-graph (type task-descriptions connectivity &rest initargs)
  (let ((graph (apply #'make-instance type initargs)))
    (setf *graph* graph)
    (set-tasks graph
      (loop for descr in task-descriptions
	  for objspace in (compute-location graph
					    task-descriptions 
					    connectivity)
	  collect (apply #'make-instance (append (first descr) (list :location objspace)))))
    (connect graph connectivity)
    graph))

(defmethod connect ((graph tig) connectivity)
  (loop for task in (tasks graph)
      for i from 0 upto (1- (length connectivity))
      do (loop for n in (aref connectivity i)
	     do (push (nth n (tasks graph))
		      (connected-tasks task))))
  graph)
		  

(defmethod set-tasks ((graph tig) (tasks list)) 
  (setf (tasks graph) tasks))

(defmethod add-task ((graph tig) (task task))
  (push task 
	(tasks graph)))






(defmethod compute-location ((strategy toy-strategy) task-descriptions connectivity
			     &optional (machine (spaces *manager*)))
  (declare (ignore connectivity))
  (loop repeat (length task-descriptions)
      for spaces = machine then (if spaces
				    (rest spaces) 
				  machine)
      collect (first spaces)))


(defmethod compute-location ((strat chaco-strategy) task-descriptions connectivity
			     &optional (machine (cons *localspace* (spaces *manager*))))
  (if (= (length machine) 1)		;there is just the local objectspace
      (make-list (length task-descriptions)
		 :initial-element *localspace*) ; return the local space as location for all tasks.
    (let ((result (make-array (length task-descriptions) 
			    :element-type 'fixnum)))
      (interface (length task-descriptions) ;nvtxs = number of vertexes
		 (start-array connectivity)
		 (adjacency-array connectivity)
		 (vertex-weights task-descriptions)
		 (edge-weights connectivity)
		 result			;assignment 
		 1			;(architecture machine)
		 0			;(hyperdim machine)
		 (make-array 3		;meshdims
			     :element-type 'fixnum
			     :initial-contents (list (length machine) 0 0))
					  ;(goal machine) 
		 2			;*global-method*
		 1			;*local-method*
		 0			;rqi_flag
		 2			;vmax
		 1			;ndims
		 (coerce 1e-3 'double-float);eigtol
		 7654321)			;seed
      (loop for x from 0 upto (1- (length task-descriptions))
	  collect (nth (aref result x) machine)))))

(defun start-array (con)
  (let ((array (make-array (1+ (length con)) :element-type 'fixnum)))
    (setf (aref array 0) 0)
    (loop for v from 1 upto (length con)
	do (setf (aref array v) 
	     (+ (aref array (1- v))
		(length (aref con (1- v))))))
    array))

(defun adjacency-array (con)
  (let* ((adj (loop for i from 0 upto (1- (length con))
		  append (mapcar #'1+ (aref con i)))))
    (make-array (length adj)
		:element-type 'fixnum
		:initial-contents adj)))

(defun vertex-weights (tasks)
  (make-array (length tasks) 
	      :element-type 'fixnum
	      :initial-contents (mapcar #'cadr tasks)))

(defun edge-weights (con)
  (let* ((weights (loop for i from 0 upto (1- (length con))
		     append (aref con i))))
    (make-array (length weights)
		:element-type 'single-float
		:initial-element 1.0)))




;;;The algorithm is an implemetation of mean-field-annealing for mapping and is explained in 
;;;Bultan, Aykanat: A New Mapping ..., Journal of P. and D. Comp. 16 (1992), 292-305.
(defmethod compute-location ((strategy mean-field-strategy) tasks connectivity
			     &optional (machine (spaces *manager*)))
  (let* ((adjacency (loop for row in connectivity
	 		collect (loop for weight in row
				    for x from 0 upto (1- (length tasks))
				    when (> weight 0)
				    collect (cons x weight))))
	  (count 0)
	  (number-of-tasks (length tasks))
	  (number-of-processors (length machine))
	 (hopfield-net (initialize (make-array (list number-of-tasks number-of-processors) 
					       :initial-element 0) 
				   strategy))
	 (adjust-value (compute-adjust hopfield-net tasks connectivity machine)))
    (loop for temp = (initial-temperature strategy) then (next-temperature strategy)
	while temp
	do (loop for i = (random number-of-tasks)
	       for taski = (nth i tasks)
	       for field = ;; instead of computing the double-sum of equation (9) again 
		 ;;and again, I compute it once without the restriction q <> p,  
		 ;; but subtract the marked term T (a few lines lower) again and again.
		 (- (loop for neighbor-node in (nth i adjacency)
			sum (loop for q from 0 upto (1- number-of-processors)
				sum ;; My Processor Graph is fully connected, 
				  ;; so there's no need for a variable for the distance
				  ;; between two processors.
				  (* (cdr neighbor-node) (aref hopfield-net (car neighbor-node) q)))))
	       for means = (loop for p from 0 upto (1- number-of-processors)
					;for space in machine
			       collect (- field 
					  (loop for j from 0 upto (1- number-of-tasks)
					      for weight in (nth i connectivity)
					      for taskj in tasks
					      unless (= j i)
					      sum (- ;; (size space) stands for the power
						   ;; of the processor. Bultan etc. dont't
						   ;; use this term, so they can't balance
						   ;; tasks on computers with divergent
						   ;; performance.
						   (* adjust-value ;(size space)
						      (cadr taskj) (cadr taski)
						      (aref hopfield-net j p))
						   ;;**** This is the term T.****
						   (* weight (aref hopfield-net j p))))))
	       for summed = (loop for mean in means
				sum (exp (/ mean temp)))
	       for spins = (loop for mean in means
			       collect (/ (exp (/ mean temp)) summed))
	       for delta = (abs (loop for mean in means
			       for spin in spins 
			       for p from 0 upto (1- number-of-processors)
			       sum (* mean (- spin (aref hopfield-net i p)))))
	       do (incf count) 
			;   (format t "Durchgang Nummer ~a: delta: ~a~% Neue spins fuer task ~a: ki1: ~a ki2:~a~% alte spins: ~a ~a~% Temperatur: ~a~%~%"
			;	 (incf count) delta i (car spins) (cdr spins) (aref hopfield-net i 0) (aref hopfield-net i 1) temp)
			
	       while (changed strategy delta)
	       do (progn  (loop for p from 0 upto (1- number-of-processors)
		      for spin in spins
			      do (setf (aref hopfield-net i p) spin)))))
  (format t "~% Count: ~a~%" count)
  (decode-location tasks machine hopfield-net)))



(defmethod size ((space list))
  (cadr space))

(defun compute-adjust (net tasks connectivity machine)
  (let ((n (1- (array-dimension net 0)))
	    (m (1- (array-dimension net 1))))
	( / (loop for i from 0 upto n
		for row in connectivity
		sum (loop for j from 0 upto n
			for weight in row
			unless (= i j)
			sum (loop for p from 0 upto m
				sum (loop for q from 0 upto m
					sum (* weight (aref net i p) (aref net j q))))))
	    (* (1+ m)
	       (loop for i from 0 upto n   
		   for taski in tasks
		   sum (loop for j from 0 upto n
			   for taskj in tasks
			   unless (= i j)
			   sum (loop for p from 0 upto m
					;for space in machine
				   sum (* (aref net i p) (aref net j p) ;(size space)
					  (cadr taski) (cadr taskj) ))))))))


(defun decode-location (tasks machine hopfield-net)
  (loop for i from 0 upto (1- (length tasks))
		       collect (loop with maxpos = ()
				   with max = 0 
				   for space in machine
				   for p from 0 upto (length machine)
				   for spin = (aref hopfield-net i p)
				   when (> spin max)
				   do (progn (setf maxpos space)
					     (setf max spin))
				   finally (return maxpos))))
   

(defun initialize (net strategy)
  (setf (current-temperature strategy) (initial-temperature strategy))
  (setf (cooling-length strategy) (array-dimension net 0))
  (loop for i from 0 upto (1- (array-dimension net 0))
	for row-sum = (loop for p from 0 upto (1- (array-dimension net 1))
			sum (setf (aref net i p) (random 1.0)))
      do (loop for p from 0 upto (1- (array-dimension net 1))
	     do (setf (aref net i p) (/ (aref net i p) row-sum))))
  net)

(defvar *end-temp* 2.5)
(defvar *fast-temp* 1.0)

(defmethod next-temperature ((strategy mean-field-strategy))
  (setf (current-temperature strategy)
    (cond  ((< (current-temperature strategy)
	       (/ (initial-temperature strategy) *end-temp*))
	    nil)
	   ((< (current-temperature strategy)
	     (/ (initial-temperature strategy) *fast-temp*))
	    (fast-cooling strategy))
	   (t (slow-cooling strategy)))))

(defmethod changed ((strategy mean-field-strategy) delta)
  (if (> delta  (epsilon strategy))
      (setf (effectless-iterations strategy) 0)
    (progn (incf (effectless-iterations strategy))
	   nil)))

(defun slow-cooling (strategy)
  (if (> (cooling-length strategy) (effectless-iterations strategy))
      (current-temperature strategy)
    (progn (setf (effectless-iterations strategy) 0)
	   (setf (current-temperature strategy)
	     (* 0.9 (current-temperature strategy))))))

(defun fast-cooling (strategy)
  (if (> (cooling-length strategy) (* 4 (effectless-iterations strategy)))
      (current-temperature strategy) 
    (progn (setf (effectless-iterations strategy) 0)
	   (setf (current-temperature strategy)
	     (* 0.5 (current-temperature strategy))))))
  

(defun evaluate (tasks machine connectivity solution)
  (+ (/ (loop for soli in solution
	    for row in connectivity
	    sum (loop for solj in solution
		    for weight in row
		    unless (equal (car soli) (car solj))
		    sum weight))
	2)
     (loop for space in machine
	 maximize (/ (loop for task in tasks
			 for sol in solution
			 when (equal space sol)
			 sum  (cadr task))
		     (size space)))))

(defun test-mean (tasks machine connectivity &optional (init-temp 5) (end-temp *end-temp*) (fast-temp *fast-temp*))
  (let ((s (make-instance 'mean-field-strategy)))
    (loop repeat 10
	for temp = init-temp then (- temp 1)
	do (setf (initial-temperature s) temp)
	do (format t "~%~%Initial Temperature:~a~%" temp)
	do (loop repeat 10
	       for *end-temp* = end-temp then ( + 0.4 *end-temp*)
	       for *fast-temp* = fast-temp then (+ 0.1 *fast-temp*)
	       for solution = (compute-location s tasks connectivity machine)
	       do (format t "end: ~a   fast: ~a evaluate: ~a   Verteilung der tasks: ~a"
			  *end-temp* *fast-temp* (evaluate tasks machine connectivity solution)
			  (loop for space in machine
			      collect (length (remove space solution :test #'equal))))))))





#|	
(setf s (make-instance 'chaco-strategy))

(setf row 20)	

(setf tasks  (loop for x from 1 upto (expt row 2)
		 collect (list (gensym "task") 1))) ;(1+ (random 9)))


;;; creates the connectivity matrix for a grid with sides of length row
(setf c2 (make-list (expt row 2)))
(loop for i from 0 upto (1- (expt row 2))
    do (let ((l (make-list (expt row 2) :initial-element 0.0)))
	 (when (> (mod i row) 0)
	   (setf (nth (1- i) l) 2.0))
	 (when (> i (1- row))
	   (setf (nth (- i row) l) 2.0))
	 (when (< i (- (expt row 2) row))
	   (setf (nth (+ row i) l) 2.0))
	  (when (< (mod i row) (1- row))
	    (setf (nth (1+ i) l) 2.0))
	  (setf (nth i c2) l)))


;;;create a random net of size row^2
(setf c (make-list (expt row 2)))
(loop for i from 0 upto (1- (expt row 2))
    do (setf (nth i c) (make-list (expt row 2) :initial-element 0)))

(loop for i from 0 upto (1- (expt row 2))
    do (loop for j from (1+ i) upto (1- (expt row 2))
	   when (< (random 20.0) 1.0)
	   do (let ((weight (random 1.0)))
		(setf (nth i (nth j c)) weight)
		(setf (nth j (nth i c)) weight))))


;;; create a machine and compute a mapping
(setf machine '(("ki1" 1) ("ki2" 1)))

(compute-location s 
		  tasks
		  c2
		  machine)
|#


